@@ -1,8 +1,16 @@
+from copy import deepcopy
 from datetime import datetime
+import json
+import os
 import os.path as op
+from pathlib import Path
 import re
 from uuid import uuid4
-from . import models
+
+import jsonschema
+
+from .dandiset import Dandiset
+from . import __version__, get_logger, models
 from .pynwb_utils import (
     _get_pynwb_metadata,
     get_neurodata_types,
@@ -10,10 +18,7 @@ from .pynwb_utils import (
     ignore_benign_pynwb_warnings,
     metadata_cache,
 )
-from .utils import ensure_datetime
-
-from . import __version__, get_logger
-from .dandiset import Dandiset
+from .utils import ensure_datetime, get_utcnow_datetime
 
 lgr = get_logger()
 
@@ -121,8 +126,6 @@ def extract_age(metadata):
         except (KeyError, TypeError, ValueError):
             return ...
     else:
-        if start < dob:
-            raise ValueError("session_start_time precedes date_of_birth")
         duration = timedelta2duration(start - dob)
     return models.PropertyValue(value=duration, unitText="Years from birth")
 
@@ -155,7 +158,7 @@ def timedelta2duration(delta):
 
 def extract_sex(metadata):
     value = metadata.get("sex", None)
-    if value is not None:
+    if value is not None and value != "":
         value = value.lower()
         if value in ["m", "male"]:
             value_id = "http://purl.obolibrary.org/obo/PATO_0000384"
@@ -163,7 +166,7 @@ def extract_sex(metadata):
         elif value in ["f", "female"]:
             value_id = "http://purl.obolibrary.org/obo/PATO_0000383"
             value = "Female"
-        elif value in ["unknown"]:
+        elif value in ["unknown", "u"]:
             value_id = None
             value = "Unknown"
         elif value in ["other"]:
@@ -181,7 +184,7 @@ def extract_sex(metadata):
 
 def extract_species(metadata):
     value = metadata.get("species", None)
-    if value is not None:
+    if value is not None and value != "":
         value = value.lower()
         if "mouse" in value or value.startswith("mus"):
             value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10090"
@@ -189,9 +192,9 @@ def extract_species(metadata):
         elif "human" in value or value.startswith("homo"):
             value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9606"
             value = "Human"
-        elif "rat" in value:
-            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10117"
-            value = "House rat"
+        elif "norvegicus" in value:
+            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10116"
+            value = "Brown rat"
         elif "mulatta" in value or "rhesus" in value:
             value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9544"
             value = "Rhesus monkey"
@@ -235,24 +238,57 @@ def extract_model(modelcls, metadata, **kwargs):
     return m
 
 
+def extract_model_list(modelcls, id_field, id_source, **kwargs):
+    def func(metadata):
+        m = extract_model(
+            modelcls, metadata, **{id_field: metadata.get(id_source)}, **kwargs
+        )
+        if all(v is None for k, v in m.dict().items() if k != "schemaKey"):
+            return []
+        else:
+            return [m]
+
+    return func
+
+
 def extract_wasDerivedFrom(metadata):
-    return [
-        extract_model(models.BioSample, metadata, identifier=metadata.get("subject_id"))
-    ]
+    derived_from = None
+    for field, sample_name in [
+        ("tissue_sample_id", "tissuesample"),
+        ("slice_id", "slice"),
+        ("cell_id", "cell"),
+    ]:
+        if metadata.get(field) is not None:
+            derived_from = [
+                models.BioSample(
+                    identifier=metadata[field],
+                    wasDerivedFrom=derived_from,
+                    sampleType=models.SampleType(name=sample_name),
+                )
+            ]
+    return derived_from
+
+
+extract_wasAttributedTo = extract_model_list(
+    models.Participant, "identifier", "subject_id"
+)
+
+extract_wasGeneratedBy = extract_model_list(
+    models.Session, "name", "session_id", identifier=None
+)
 
 
 def extract_digest(metadata):
     if "digest" in metadata:
-        return models.Digest(
-            value=metadata["digest"],
-            cryptoType=models.DigestType[metadata["digest_type"]],
-        )
+        return {models.DigestType[metadata["digest_type"]]: metadata["digest"]}
     else:
         return ...
 
 
 FIELD_EXTRACTORS = {
     "wasDerivedFrom": extract_wasDerivedFrom,
+    "wasAttributedTo": extract_wasAttributedTo,
+    "wasGeneratedBy": extract_wasGeneratedBy,
     "age": extract_age,
     "sex": extract_sex,
     "assayType": extract_assay_type,
@@ -269,7 +305,15 @@ def extract_field(field, metadata):
         return metadata.get(field, ...)
 
 
-def nwb2asset(nwb_path, digest=None, digest_type=None):
+def nwb2asset(
+    nwb_path, digest=None, digest_type=None, schema_version=None
+) -> models.BareAssetMeta:
+    if schema_version is not None:
+        current_version = models.get_schema_version()
+        if schema_version != current_version:
+            raise ValueError(
+                f"Unsupported schema version: {schema_version}; expected {current_version}"
+            )
     start_time = datetime.now().astimezone()
     metadata = get_metadata(nwb_path)
     if digest is not None:
@@ -277,27 +321,62 @@ def nwb2asset(nwb_path, digest=None, digest_type=None):
         metadata["digest_type"] = digest_type
     metadata["contentSize"] = op.getsize(nwb_path)
     metadata["encodingFormat"] = "application/x-nwb"
+    metadata["dateModified"] = get_utcnow_datetime()
+    metadata["blobDateModified"] = ensure_datetime(os.stat(nwb_path).st_mtime)
+    if metadata["blobDateModified"] > metadata["dateModified"]:
+        lgr.warning(
+            "mtime %s of %s is in the future", metadata["blobDateModified"], nwb_path
+        )
     asset = metadata2asset(metadata)
     end_time = datetime.now().astimezone()
-    asset.wasGeneratedBy = models.Activity(
-        identifier=str(uuid4()),
-        name="Metadata generation",
-        description="Metadata generated by DANDI cli",
-        wasAssociatedWith=models.Software(
-            identifier={"propertyID": "RRID", "value": "SCR_019009"},
-            name="DANDI Command Line Interface",
-            description=f"dandi-cli {__version__}",
-            version=__version__,
-            url="https://github.com/dandi/dandi-cli",
-        ),
-        startedAt=start_time,
-        endedAt=end_time,
-    )
+    if asset.wasGeneratedBy is None:
+        asset.wasGeneratedBy = []
+    asset.wasGeneratedBy.append(get_generator(start_time, end_time))
     return asset
 
 
+def get_default_metadata(path, digest=None, digest_type=None) -> models.BareAssetMeta:
+    start_time = datetime.now().astimezone()
+    if digest is not None:
+        digest_model = {models.DigestType[digest_type]: digest}
+    else:
+        digest_model = []
+    dateModified = get_utcnow_datetime()
+    blobDateModified = ensure_datetime(os.stat(path).st_mtime)
+    if blobDateModified > dateModified:
+        lgr.warning("mtime %s of %s is in the future", blobDateModified, path)
+    end_time = datetime.now().astimezone()
+    return models.BareAssetMeta.unvalidated(
+        contentSize=os.path.getsize(path),
+        digest=digest_model,
+        dateModified=dateModified,
+        blobDateModified=blobDateModified,
+        wasGeneratedBy=[get_generator(start_time, end_time)],
+        # encodingFormat # TODO
+    )
+
+
+def get_generator(start_time: datetime, end_time: datetime) -> models.Activity:
+    return models.Activity(
+        identifier=str(uuid4()),
+        name="Metadata generation",
+        description="Metadata generated by DANDI cli",
+        wasAssociatedWith=[
+            models.Software(
+                identifier="RRID:SCR_019009",
+                name="DANDI Command Line Interface",
+                version=__version__,
+                url="https://github.com/dandi/dandi-cli",
+                schemaKey="Software",
+            )
+        ],
+        startedAt=start_time,
+        endedAt=end_time,
+    )
+
+
 def metadata2asset(metadata):
-    return extract_model(models.AssetMeta, metadata)
+    return extract_model(models.BareAssetMeta, metadata)
 
 
 """
@@ -331,11 +410,13 @@ mapping = {
 }
 
 
-def toContributor(value):
+def toContributor(value, contrib_type):
     if not isinstance(value, list):
         value = [value]
     out = []
     for item in value:
+        if item == {"orcid": "", "roles": []}:
+            continue
         contrib = {}
         if "name" in item:
             name = item["name"].split()
@@ -348,29 +429,40 @@ def toContributor(value):
                     roles.append("".join([val.capitalize() for val in tmp]))
                 else:
                     roles.append(tmp.pop())
-            contrib["roleName"] = roles
+            contrib["roleName"] = [getattr(models.RoleType, role) for role in roles]
             del item["roles"]
+        elif contrib_type == "sponsors":
+            contrib["roleName"] = [models.RoleType.Funder]
         if "awardNumber" in item:
             contrib["awardNumber"] = item["awardNumber"]
             del item["awardNumber"]
         if "orcid" in item:
             if item["orcid"]:
-                contrib["identifier"] = models.PropertyValue(
-                    value=item["orcid"], propertyID="ORCID"
-                )
-            else:
-                contrib["identifier"] = models.PropertyValue()
+                contrib["identifier"] = item["orcid"]
+            # else:
+            #    contrib["identifier"] = models.PropertyValue()
             del item["orcid"]
+        if "affiliation" in item:
+            item["affiliation"] = [models.Organization(name=item["affiliation"])]
         if "affiliations" in item:
-            item["affiliation"] = item["affiliations"]
+            item["affiliation"] = [
+                models.Organization(name=affiliate)
+                for affiliate in item["affiliations"]
+            ]
             del item["affiliations"]
         contrib.update(**{f"{k}": v for k, v in item.items()})
+        if "awardNumber" in contrib or contrib_type == "sponsors":
+            contrib = models.Organization(**contrib)
+        else:
+            if "name" not in contrib:
+                contrib["name"] = "Last, First"
+            contrib = models.Person(**contrib)
         out.append(contrib)
     return out
 
 
 def convertv1(data):
-    oldmeta = data["dandiset"] if "dandiset" in data else data
+    oldmeta = deepcopy(data["dandiset"]) if "dandiset" in data else deepcopy(data)
     newmeta = {}
     for oldkey, value in oldmeta.items():
         if oldkey in ["language", "altid", "number_of_slices"]:
@@ -382,15 +474,19 @@ def convertv1(data):
         else:
             newkey = mapping[oldkey][0]
         if oldkey in ["contributors", "sponsors"]:
-            value = toContributor(value)
+            value = toContributor(value, oldkey)
         if oldkey == "access":
             value = [
                 models.AccessRequirements(
                     status=models.AccessType.Open, email=value["access_contact_email"]
                 )
             ]
+        if oldkey == "license":
+            value = [
+                getattr(models.LicenseType, value.replace("-", "").replace(".", ""))
+            ]
         if oldkey == "identifier":
-            value = models.PropertyValue(value=value, propertyID="DANDI")
+            value = f"DANDI:{value}"
         if len(mapping[oldkey]) == 2:
             extra = mapping[oldkey][1]
             if newkey == "contributor":
@@ -404,14 +500,23 @@ def convertv1(data):
                 out = []
                 for item in value:
                     if isinstance(item, dict):
-                        out.append({k: v for k, v in item.items()})
-                    else:
-                        present = False
-                        for val in out:
-                            if item in val.values():
-                                present = True
-                        if not present:
-                            out.append({"url": item})
+                        if (
+                            "relation" in item
+                            and "publication" in item["relation"].lower()
+                        ):
+                            del item["relation"]
+                        if "relation" not in item:
+                            if oldkey == "publications":
+                                item["relation"] = models.RelationType.IsDescribedBy
+                            if oldkey == "associatedData":
+                                item["relation"] = models.RelationType.IsDerivedFrom
+                        out.append(models.Resource(**item))
+                    elif not any(item in val.dict().values() for val in out):
+                        out.append(
+                            models.Resource(
+                                url=item, relation=models.RelationType.IsDescribedBy
+                            )
+                        )
                 value = out
             if oldkey in [
                 "number_of_subjects",
@@ -423,7 +528,14 @@ def convertv1(data):
             if isinstance(value, list):
                 for val in value:
                     if extrakey:
-                        val[extrakey] = extra
+                        if extrakey == "relation":
+                            val.relation = getattr(models.RelationType, extra)
+                        elif extrakey == "roleName":
+                            val.roleName = [
+                                getattr(models.RoleType, role) for role in extra
+                            ]
+                        else:
+                            val[extrakey] = extra
             if isinstance(value, dict):
                 if extrakey:
                     value[extrakey] = extra
@@ -433,7 +545,7 @@ def convertv1(data):
                 if oldkey == "sex":
                     vm["value"] = value
                 else:
-                    if "maximum" in value:
+                    if "maximum" in value and isinstance(value["maximum"], str):
                         if "days" in value["maximum"]:
                             value["units"] = "days"
                         if "Gestational" in value["maximum"]:
@@ -444,7 +556,7 @@ def convertv1(data):
                             value["units"] = value["maximum"][-1]
                         if "None" not in value["maximum"]:
                             value["maximum"] = float(value["maximum"].split()[0])
-                    if "minimum" in value:
+                    if "minimum" in value and isinstance(value["minimum"], str):
                         if "days" in value["minimum"]:
                             value["units"] = "days"
                         if "Gestational" in value["minimum"]:
@@ -465,6 +577,14 @@ def convertv1(data):
                         newvalues.append(val["species"])
                 vm = {"name": "species", "value": newvalues}
             value = vm
+        if newkey == "about":
+            newvalues = []
+            for item in value:
+                id = item.get("identifier", None)
+                if id is not None and not id.startswith("http"):
+                    item["identifier"] = None
+                newvalues.append(models.TypeModel(**item))
+            value = newvalues
         if newkey not in newmeta:
             newmeta[newkey] = value
         else:
@@ -483,5 +603,81 @@ def convertv1(data):
 
 def migrate2newschema(meta):
     newmeta = convertv1(meta)
-    dandimeta = models.DandiMeta.unvalidated(**newmeta)
+    dandimeta = models.DandisetMeta.unvalidated(**newmeta)
     return dandimeta
+
+
+def generate_context():
+    import pydantic
+
+    fields = {
+        "@version": 1.1,
+        "dandi": "http://schema.dandiarchive.org/",
+        "DANDI": "http://identifiers.org/DANDI:",
+        "dct": "http://purl.org/dc/terms/",
+        "owl": "http://www.w3.org/2002/07/owl#",
+        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
+        "rdfa": "http://www.w3.org/ns/rdfa#",
+        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
+        "schema": "http://schema.org/",
+        "xsd": "http://www.w3.org/2001/XMLSchema#",
+        "skos": "http://www.w3.org/2004/02/skos/core#",
+        "prov": "http://www.w3.org/ns/prov#",
+        "pav": "http://purl.org/pav/",
+        "nidm": "http://purl.org/nidash/nidm#",
+        "uuid": "http://uuid.repronim.org/",
+        "rs": "http://schema.repronim.org/",
+        "RRID": "https://scicrunch.org/resolver/RRID:",
+        "ORCID": "https://orcid.org/",
+        "ROR": "https://ror.org/",
+        "PATO": "http://purl.obolibrary.org/obo/PATO_",
+    }
+    for val in dir(models):
+        klass = getattr(models, val)
+        if not isinstance(klass, pydantic.main.ModelMetaclass):
+            continue
+        if hasattr(klass, "_ldmeta"):
+            if "nskey" in klass._ldmeta:
+                name = klass.__name__
+                fields[name] = f'{klass._ldmeta["nskey"]}:{name}'
+        for name, field in klass.__fields__.items():
+            if name == "identifier":
+                fields[name] = "@id"
+            elif name == "schemaKey":
+                fields[name] = "@type"
+            elif name == "digest":
+                fields[name] = "@nest"
+            elif "nskey" in field.field_info.extra:
+                if name not in fields:
+                    fields[name] = {"@id": field.field_info.extra["nskey"] + ":" + name}
+                    if "List" in str(field.outer_type_):
+                        fields[name]["@container"] = "@set"
+                    if name == "contributor":
+                        fields[name]["@container"] = "@list"
+                    if "enum" in str(field.type_) or name == "url":
+                        fields[name]["@type"] = "@id"
+    for item in models.DigestType:
+        fields[item.value] = {"@id": item.value, "@nest": "digest"}
+    return {"@context": fields}
+
+
+def publish_model_schemata(releasedir):
+    version = models.get_schema_version()
+    vdir = Path(releasedir, version)
+    vdir.mkdir(exist_ok=True, parents=True)
+    (vdir / "dandiset.json").write_text(models.DandisetMeta.schema_json(indent=2))
+    (vdir / "asset.json").write_text(models.AssetMeta.schema_json(indent=2))
+    (vdir / "context.json").write_text(json.dumps(generate_context(), indent=2))
+    return vdir
+
+
+def validate_dandiset_json(data, schema_dir):
+    with Path(schema_dir, "dandiset.json").open() as fp:
+        schema = json.load(fp)
+    jsonschema.validate(data, schema)
+
+
+def validate_asset_json(data, schema_dir):
+    with Path(schema_dir, "asset.json").open() as fp:
+        schema = json.load(fp)
+    jsonschema.validate(data, schema)
